from abc import ABCMeta, abstractmethod

from .IAppRegistration import IAppRegistration
from .ICategorizableResource import ICategorizableResource
from .IWebAppAssociatedResource import IWebAppAssociatedResource

class IKeyVault(ICategorizableResource, IWebAppAssociatedResource):
    """An abstract class that defines properties for any kind of Azure Key Vault"""

    __metaclass__ = ABCMeta
    
    ACCESS_KEY = "access"
    """The key for the access property"""

    @property
    @abstractmethod
    def access(self) -> list[IAppRegistration]:
        """The App Registrations that have access to the Key Vault"""

        pass

    @abstractmethod
    def app_has_access(self, app_object_id: str) -> bool:
        """Check if an App Registration has access to this Key Vault

        Args:
            app_object_id (str): The App Registration's object ID

        Returns:
            bool: True if the App Registration has access, False otherwise
        """

        pass


from typing import Any

from .interfaces.IAppRegistration import IAppRegistration

class AppRegistration(IAppRegistration):
    def __init__(self, definition_json: dict[str, Any]):
        self.__definition_json = definition_json

        self.__name = definition_json[self.NAME_KEY]
        
        self._app_id = None
        self._app_secret = None
        self._service_principal_id = None
    
    @property
    def name(self) -> str:
        """The name of the app registration"""
        
        return self.__name
    
    @property
    def app_id(self) -> str:
        """The app id of the app registration"""
        
        return self._app_id # type: ignore
    
    @app_id.setter
    def app_id(self, value: str):
        self._app_id = value
    
    @property
    def app_secret(self) -> str:
        """The app secret of associated with the app registration"""
        
        return self._app_secret # type: ignore
    
    @app_secret.setter
    def app_secret(self, value: str):
        self._app_secret = value
    
    @property
    def service_principal_id(self) -> str:
        """The id of the service principal associated with the app registration"""
        
        return self._service_principal_id # type: ignore
    
    @service_principal_id.setter
    def service_principal_id(self, value: str):
        self._service_principal_id = value
    
    def to_json(self) -> dict[str, Any]:
        return_obj = {
            self.NAME_KEY: self.name
        }

        if self.app_id is not None:
            return_obj["appId"] = self.app_id
        
        if self.app_secret is not None:
            return_obj["appSecret"] = self.app_secret
        
        if self.service_principal_id is not None:
            return_obj["servicePrincipalId"] = self.service_principal_id
        
        return return_obj


from typing import Any, Literal

from .interfaces.IAzureResources import IAzureResources

from .ResourceGroup import ResourceGroup
from .WebApp import WebApp
from .ContainerRegistry import ContainerRegistry
from .AppRegistration import AppRegistration
from .Database import Database
from .RedisCache import RedisCache
from .StorageAccount import StorageAccount
from .KeyVault import KeyVault

class AzureResources(IAzureResources):
    def __init__(self, definition_json: dict[str, Any]):
        self.__definition_json = definition_json
        
        # Create resource group objects from the definition JSON and store them in a list
        self.__resource_groups: list[ResourceGroup] = []
        for resource_group_definition in definition_json[IAzureResources.RESOURCE_GROUPS_KEY]:
            resource_group = ResourceGroup(resource_group_definition)
            self.__resource_groups.append(resource_group)
        
        # Create webapp objects from the definition JSON and store them in a list
        self.__web_apps: list[WebApp] = []
        for web_app_definition in definition_json[IAzureResources.WEB_APPS_KEY]:
            web_app = WebApp(web_app_definition, self)
            self.__web_apps.append(web_app)
        
        # Create container registry objects from the definition JSON and store them in a list
        self.__container_registries: list[ContainerRegistry] = []
        if IAzureResources.CONTAINER_REGISTRIES_KEY in definition_json:
            for container_registry_definition in definition_json[IAzureResources.CONTAINER_REGISTRIES_KEY]:
                container_registry = ContainerRegistry(container_registry_definition, self)
                self.__container_registries.append(container_registry)
        
        # Create app registration objects from the definition JSON and store them in a list
        self.__app_registrations: list[AppRegistration] = []
        if IAzureResources.APP_REGISTRATIONS_KEY in definition_json:
            for app_registration_definition in definition_json[IAzureResources.APP_REGISTRATIONS_KEY]:
                app_registration = AppRegistration(app_registration_definition)
                self.__app_registrations.append(app_registration)
        
        self.__databases: list[Database] = []
        if IAzureResources.DATABASES_KEY in definition_json:
            for database_definition in definition_json[IAzureResources.DATABASES_KEY]:
                database = Database(database_definition, self)
                self.__databases.append(database)
        
        self.__redis_caches: list[RedisCache] = []
        if IAzureResources.CACHES_KEY in definition_json:
            for redis_cache_definition in definition_json[IAzureResources.CACHES_KEY]:
                redis_cache = RedisCache(redis_cache_definition, self)
                self.__redis_caches.append(redis_cache)

        self.__storage_accounts: list[StorageAccount] = []
        if IAzureResources.STORAGE_KEY in definition_json:
            for storage_account_definition in definition_json[IAzureResources.STORAGE_KEY]:
                storage_account = StorageAccount(storage_account_definition, self)
                self.__storage_accounts.append(storage_account)
        
        self.__key_vaults: list[KeyVault] = []
        if IAzureResources.VAULTS_KEY in definition_json:
            for key_vault_definition in definition_json[IAzureResources.VAULTS_KEY]:
                key_vault = KeyVault(key_vault_definition, self)
                self.__key_vaults.append(key_vault)
        
        # Create categorized resources objects from the different category resources defined in the definition JSON and store 
        # them in a dictionary (by category name)
        self.__categorized_resources = self.__get_categorized_resources(definition_json)

    # =================
    # Internal Methods
    # =================

    def __get_categorized_resources(self, definition_json: dict[str, Any]) -> dict[str, list[ContainerRegistry | Database | StorageAccount | KeyVault | RedisCache | WebApp]]:
        """Gets the categorized resources from the definition JSON

        Args:
            definition_json (dict): The definition JSON

        Returns:
            dict[str, list[ContainerRegistry | Database | StorageAccount | KeyVault | RedisCache | WebApp]]: The categorized resources dictionary (by category name)
        """

        # Create categorized resources objects from the different category resources defined in the definition JSON and store 
        # them in a dictionary (by category name)
        categorized_resources: dict[str, list[ContainerRegistry | Database | StorageAccount | KeyVault | RedisCache | WebApp]] = {}

        if IAzureResources.CATEGORIZED_RESOURCES_KEY not in definition_json:
            return categorized_resources
        
        for category_name, category_resources_by_type in dict(definition_json[self.CATEGORIZED_RESOURCES_KEY]).items():
            # Variable to hold the list of resources for the current category
            category_resources: list[ContainerRegistry | Database | StorageAccount | KeyVault | RedisCache | WebApp] = []
            
            if IAzureResources.CONTAINER_REGISTRIES_KEY in category_resources_by_type:
                for container_registry_definition in category_resources_by_type[IAzureResources.CONTAINER_REGISTRIES_KEY]:
                    container_registry = ContainerRegistry(container_registry_definition, self, category_name)
                    category_resources.append(container_registry)
            
            if IAzureResources.DATABASES_KEY in category_resources_by_type:
                for database_definition in category_resources_by_type[IAzureResources.DATABASES_KEY]:
                    database = Database(database_definition, self, category_name)
                    category_resources.append(database)
            
            if IAzureResources.STORAGE_KEY in category_resources_by_type:
                for storage_account_definition in category_resources_by_type[IAzureResources.STORAGE_KEY]:
                    storage_account = StorageAccount(storage_account_definition, self, category_name)
                    category_resources.append(storage_account)
            
            if IAzureResources.VAULTS_KEY in category_resources_by_type:
                for key_vault_definition in category_resources_by_type[IAzureResources.VAULTS_KEY]:
                    key_vault = KeyVault(key_vault_definition, self, category_name)
                    category_resources.append(key_vault)
            
            if IAzureResources.CACHES_KEY in category_resources_by_type:
                for redis_cache_definition in category_resources_by_type[IAzureResources.CACHES_KEY]:
                    redis_cache = RedisCache(redis_cache_definition, self, category_name)
                    category_resources.append(redis_cache)
            
            if IAzureResources.WEB_APPS_KEY in category_resources_by_type:
                for web_app_definition in category_resources_by_type[IAzureResources.WEB_APPS_KEY]:
                    web_app = WebApp(web_app_definition, self, category_name)
                    category_resources.append(web_app)

            categorized_resources[category_name] = category_resources
        
        return categorized_resources

    # ================
    # Resource Groups
    # ================

    @property
    def resource_groups(self) -> list[ResourceGroup]: # type: ignore
        """The resource groups in the Azure Resources"""

        return self.__resource_groups
    
    def add_resource_group(self, resource_group: ResourceGroup):
        """Add a resource group to the Azure Resources

        Args:
            resource_group (ResourceGroup): The resource group to add
        """

        self.__resource_groups.append(resource_group)

    def has_resource_group(self, name: str) -> bool:
        """Check if the Azure Resources has a resource group with the specified name
        
        Args:
            name (str): The name of the resource group to check for
        
        Returns:
            bool: True if the Azure Resources has a resource group with the specified name, False otherwise
        """

        # Loop over each resource group in the Azure Resources and check if the name matches
        for resource_group in self.resource_groups:
            if resource_group.name == name:
                return True
        
        return False
    
    def get_resource_group(self, name: str) -> ResourceGroup | None:
        """Get the resource group with the specified name

        Args:
            name (str): The name of the resource group to get

        Returns:
            dict | None: The resource group with the specified name or None if it does not exist
        """
        
        if self.has_resource_group(name):
            for resource_group in self.resource_groups:
                if resource_group.name == name:
                    return resource_group
        
        return None
    
    # =========
    # Web Apps
    # =========

    @property
    def web_apps(self) -> list[WebApp]: # type: ignore
        """The web apps in the Azure Resources"""

        return self.__web_apps
    
    def add_web_apps(self, web_app: WebApp): # type: ignore
        """Add a web app to the Azure Resources

        Args:
            web_app (WebApp): The web app to add
        """

        self.__web_apps.append(web_app)

    def has_web_app(self, name: str) -> bool:
        """Check if the Azure Resources has a web app with the specified name
        
        Args:
            name (str): The name of the web app to check for
        
        Returns:
            bool: True if the Azure Resources has a web app with the specified name, False otherwise
        """

        for web_app in self.web_apps:
            if web_app.name == name:
                return True
            
        return False
    
    def get_web_app(self, name: str) -> WebApp | None:
        """Get the web app with the specified name

        Args:
            name (str): The name of the web app to get

        Returns:
            dict | None: The web app with the specified name or None if it does not exist
        """
        
        if self.has_web_app(name):
            for web_app in self.web_apps:
                if web_app.name == name:
                    return web_app
        
        return None

    # =====================
    # Container Registries
    # =====================

    @property
    def container_registries(self) -> list[ContainerRegistry]: # type: ignore
        """The container registries in the Azure Resources"""

        return self.__container_registries
    
    def add_container_registry(self, container_registry: ContainerRegistry): # type: ignore
        """Add a container registry to the Azure Resources

        Args:
            container_registry (ContainerRegistry): The container registry to add
        """

        self.__container_registries.append(container_registry)

    def has_container_registry(self, name: str) -> bool:
        """Check if the Azure Resources has a container registry with the specified name
        
        Args:
            name (str): The name of the container registry to check for
        
        Returns:
            bool: True if the Azure Resources has a container registry with the specified name, False otherwise
        """

        for container_registry in self.container_registries:
            if container_registry.name == name:
                return True
            
        return False
    
    def get_container_registry(self, name: str) -> ContainerRegistry | None:
        """Get the container registry with the specified name

        Args:
            name (str): The name of the container registry to get

        Returns:
            dict | None: The container registry with the specified name or None if it does not exist
        """
        
        if self.has_container_registry(name):
            for container_registry in self.container_registries:
                if container_registry.name == name:
                    return container_registry
        
        return None
    
    # ==================
    # App Registrations
    # ==================

    @property
    def app_registrations(self) -> list[AppRegistration]: # type: ignore
        """The app registrations in the Azure Resources"""

        return self.__app_registrations
    
    def add_app_registration(self, app_registration: AppRegistration):
        """Add an app registration to the Azure Resources

        Args:
            app_registration (AppRegistration): The app registration to add
        """

        self.__app_registrations.append(app_registration)

    def has_app_registration(self, name: str) -> bool:
        """Check if the Azure Resources has an app registration with the specified name
        
        Args:
            name (str): The name of the app registration to check for
        
        Returns:
            bool: True if the Azure Resources has an app registration with the specified name, False otherwise
        """

        for app_registration in self.app_registrations:
            if app_registration.name == name:
                return True
            
        return False
    
    def get_app_registration(self, name: str) -> AppRegistration | None:
        """Get the app registration with the specified name

        Args:
            name (str): The name of the app registration to get

        Returns:
            dict | None: The app registration with the specified name or None if it does not exist
        """
        
        if self.has_app_registration(name):
            for app_registration in self.app_registrations:
                if app_registration.name == name:
                    return app_registration
        
        return None
    
    # ==========
    # Databases
    # ==========

    @property
    def databases(self) -> list[Database]: # type: ignore
        """The databases in the Azure Resources"""

        return self.__databases
    
    def add_database(self, database: Database): # type: ignore
        """Add a database to the Azure Resources

        Args:
            database (Database): The database to add
        """

        self.__databases.append(database)

    def has_database(self, name: str) -> bool:
        """Check if the Azure Resources has a database with the specified name
        
        Args:
            name (str): The name of the database to check for
        
        Returns:
            bool: True if the Azure Resources has a database with the specified name, False otherwise
        """

        for database in self.databases:
            if database.name == name:
                return True
            
        return False
    
    def get_database(self, name: str) -> Database | None:
        """Get the database with the specified name

        Args:
            name (str): The name of the database to get

        Returns:
            dict | None: The database with the specified name or None if it does not exist
        """
        
        if self.has_database(name):
            for database in self.databases:
                if database.name == name:
                    return database
        
        return None

    # =======
    # Caches
    # =======

    @property
    def caches(self) -> list[RedisCache]: # type: ignore
        """The caches in the Azure Resources"""

        return self.__redis_caches
    
    def add_cache(self, cache: RedisCache): # type: ignore
        """Add a cache to the Azure Resources

        Args:
            cache (RedisCache): The cache to add
        """

        self.__redis_caches.append(cache) 

    def has_cache(self, name: str) -> bool:
        """Check if the Azure Resources has a cache with the specified name
        
        Args:
            name (str): The name of the cache to check for
        
        Returns:
            bool: True if the Azure Resources has a cache with the specified name, False otherwise
        """

        for cache in self.caches:
            if cache.name == name:
                return True
            
        return False
    
    def get_cache(self, name: str) -> RedisCache | None:
        """Get the cache with the specified name

        Args:
            name (str): The name of the cache to get

        Returns:
            dict | None: The cache with the specified name or None if it does not exist
        """
        
        if self.has_cache(name):
            for cache in self.caches:
                if cache.name == name:
                    return cache
        
        return None
    
    # ===============================
    # Uncategorized Storage Accounts
    # ===============================

    @property
    def uncategorized_storage(self) -> list[StorageAccount]: # type: ignore
        """The uncategorized storage accounts in the Azure Resources"""

        return self.__storage_accounts
    
    def add_uncategorized_storage(self, storage: StorageAccount): # type: ignore
        """Add an uncategorized storage account to the Azure Resources

        Args:
            storage (StorageAccount): The uncategorized storage account to add
        """

        self.__storage_accounts.append(storage)

    def has_uncategorized_storage(self, name: str) -> bool:
        """Check if the Azure Resources has an uncategorized storage account with the specified name
        
        Args:
            name (str): The name of the uncategorized storage account to check for
        
        Returns:
            bool: True if the Azure Resources has an uncategorized storage account with the specified name, False otherwise
        """

        for storage in self.uncategorized_storage:
            if storage.name == name:
                return True
            
        return False
    
    def get_uncategorized_storage(self, name: str) -> StorageAccount | None:
        """Get the uncategorized storage account with the specified name

        Args:
            name (str): The name of the uncategorized storage account to get

        Returns:
            dict | None: The uncategorized storage account with the specified name or None if it does not exist
        """
        
        if self.has_uncategorized_storage(name):
            for storage in self.uncategorized_storage:
                if storage.name == name:
                    return storage
        
        return None
    
    # =========================
    # Uncategorized Key Vaults
    # =========================

    @property
    def uncategorized_vaults(self) -> list[KeyVault]: # type: ignore
        """The uncategorized key vaults in the Azure Resources"""

        return self.__key_vaults
    
    def add_uncategorized_vaults(self, vault: KeyVault): # type: ignore
        """Add the uncategorized key vault to the Azure Resources

        Args:
            vault (KeyVault): The uncategorized key vault to add to the Azure Resources
        """

        self.__key_vaults.append(vault)

    def has_uncategorized_vault(self, name: str) -> bool:
        """Check if the Azure Resources has an uncategorized key vault with the specified name
        
        Args:
            name (str): The name of the uncategorized key vault to check for
        
        Returns:
            bool: True if the Azure Resources has an uncategorized key vault with the specified name, False otherwise
        """

        for vault in self.uncategorized_vaults:
            if vault.name == name:
                return True
            
        return False
    
    def get_uncategorized_vault(self, name: str) -> KeyVault | None:
        """Get the uncategorized key vault with the specified name

        Args:
            name (str): The name of the uncategorized key vault to get

        Returns:
            dict | None: The uncategorized key vault with the specified name or None if it does not exist
        """
        
        if self.has_uncategorized_vault(name):
            for vault in self.uncategorized_vaults:
                if vault.name == name:
                    return vault
        
        return None

    # =====================================================
    # Categorized Resources (BA Infrastructure Categories)
    # =====================================================

    @property
    def categorized_resources(self) -> dict[str, list[ContainerRegistry | Database | StorageAccount | KeyVault | RedisCache | WebApp]]: # type: ignore
        """The categorized resources in the Azure Resources"""
        
        return self.__categorized_resources
    
    def add_categorized_resource(self, category: str, *resource: ContainerRegistry | Database | StorageAccount | KeyVault | RedisCache | WebApp): # type: ignore
        """Add a categorized resource to the Azure Resources

        Args:
            category (str): The category of the resource to add
            resource (ContainerRegistry | Database | StorageAccount | KeyVault | RedisCache | WebApp): The resource to add
        """

        if category not in self.categorized_resources:
            self.categorized_resources[category] = []
        
        for res in resource:
            self.categorized_resources[category].append(res)

    def get_resource_categories(self) -> list[str]:
        """Get the categories of the resources in the Azure Resources

        Returns:
            list[str]: The categories of the resources in the Azure Resources
        """

        return list(self.categorized_resources.keys())
    
    def has_resource_category(self, category: str) -> bool:
        """Check if the Azure Resources has a resource category with the specified name
        
        Args:
            category (str): The category of the resource to check for
        
        Returns:
            bool: True if the Azure Resources has a resource category with the specified name, False otherwise
        """

        return category in self.categorized_resources
    
    def get_category_resources(self, category: str, types: list[Literal['ContainerRegistry'] | Literal['Database'] | Literal['StorageAccount'] | Literal['KeyVault'] | Literal['RedisCache'] | Literal['WebApp']] = ['ContainerRegistry', 'Database', 'StorageAccount', 'KeyVault', 'RedisCache', 'WebApp']) -> list[ContainerRegistry | Database | StorageAccount | KeyVault | RedisCache | WebApp] | None: # type: ignore
        """Get the resources in the Azure Resources with the specified category
        
        Args:
            category (str): The category of the resources to get
            types (list['ContainerRegistry' | 'Database' | 'StorageAccount' | 'KeyVault' | 'RedisCache' | 'WebApp']): The types of resources to get (default: all types)
        
        Returns:
            CategorizedResources | None: The resources in the Azure Resources with the specified category or None if the category does not exist
        """

        if self.has_resource_category(category):
            category_resources = self.categorized_resources[category]
            return_category_resources: list[ContainerRegistry | Database | StorageAccount | KeyVault | RedisCache | WebApp] = []
            
            for resource in category_resources:
                if type(resource).__name__ in types:
                    return_category_resources.append(resource)
            
            return return_category_resources
        
        return None
    
    # ==============
    # Other methods
    # ==============

    def to_json(self) -> dict[str, Any]:
        output = {
            self.RESOURCE_GROUPS_KEY: [resource_group.to_json() for resource_group in self.resource_groups],
            self.WEB_APPS_KEY: [web_app.to_json() for web_app in self.web_apps],
            self.CONTAINER_REGISTRIES_KEY: [container_registry.to_json() for container_registry in self.container_registries],
            self.APP_REGISTRATIONS_KEY: [app_registration.to_json() for app_registration in self.app_registrations],
            self.DATABASES_KEY: [database.to_json() for database in self.databases],
            self.CACHES_KEY: [cache.to_json() for cache in self.caches],
            self.STORAGE_KEY: [storage_account.to_json() for storage_account in self.uncategorized_storage],
            self.VAULTS_KEY: [vault.to_json() for vault in self.uncategorized_vaults],
        }

        categorized_resources = {}
        for category in self.get_resource_categories():
            categorized_resources[category] = {}
            
            category_storage_accounts: list[StorageAccount] = self.get_category_resources(category, ['StorageAccount']) # type: ignore
            categorized_resources[category][self.STORAGE_KEY] = [categorized_storage_account.to_json() for categorized_storage_account in category_storage_accounts]

            category_vaults: list[KeyVault] = self.get_category_resources(category, ['KeyVault']) # type: ignore
            categorized_resources[category][self.VAULTS_KEY] = [categorized_vault.to_json() for categorized_vault in category_vaults]
        
        output[self.CATEGORIZED_RESOURCES_KEY] = categorized_resources # type: ignore

        return output


from typing import Any

from .interfaces.IStorageAccount import IStorageAccount
from .interfaces.IAzureResources import IAzureResources

from .ResourceGroup import ResourceGroup
from .WebApp import WebApp
from .StorageContainer import StorageContainer
from .StorageTable import StorageTable

class StorageAccount(IStorageAccount):
    def __init__(self, definition_json: dict[str, Any], azure_resources: IAzureResources, category_name: str | None = None):
        self.__definition_json = definition_json

        self.__name = definition_json[self.NAME_KEY]
        self.__category_name = category_name
        self.__key = ''
        self.__connection_string = ''
        self.__web_apps = self.__get_web_apps(definition_json, azure_resources)
        self.__resource_group = self.__get_resource_group(definition_json, azure_resources)
        self.__containers = self.__get_containers(definition_json)
        self.__tables = self.__get_tables(definition_json)
    
    def __get_resource_group(self, definition_json: dict[str, Any], azure_resources: IAzureResources) -> ResourceGroup:
        """Gets the resource group from the definition json

        Args:
            definition_json (dict): The definition json
            azure_resources (IAzureResources): The azure resources (used to get the resource group)

        Raises:
            Exception: If the resource group is not found

        Returns:
            ResourceGroup: The resource group
        """

        resource_group: ResourceGroup | None = azure_resources.get_resource_group(definition_json[self.RESOURCE_GROUP_KEY]) if self.RESOURCE_GROUP_KEY in definition_json else None # type: ignore
        if resource_group is None and len(self.__web_apps) > 0:
            web_app_rgs: list[ResourceGroup] = [web_app.resource_group for web_app in self.__web_apps if web_app.resource_group]
            if len(web_app_rgs) == 1:
                resource_group = web_app_rgs[0]
        
        if resource_group is None:
            raise Exception(f"Could not find resource group for storage account {self.name}")
    
        return resource_group

    def __get_web_apps(self, definition_json: dict[str, Any], azure_resources: IAzureResources) -> list[WebApp]:
        """Gets the associated web apps from the definition json

        Args:
            definition_json (dict): The definition json
            azure_resources (IAzureResources): The azure resources (used to get the web apps)

        Returns:
            list[WebApp]: The list associated web apps
        """

        if self.WEB_APPS_KEY in definition_json and type(definition_json[self.WEB_APPS_KEY]) is list:
            web_apps: list[WebApp] = []
            
            for web_app in definition_json[self.WEB_APPS_KEY]:
                web_app_object: WebApp = azure_resources.get_web_app(str(web_app)) # type: ignore
                web_apps.append(web_app_object)
            
            return web_apps
        else:
            return []

    def __get_containers(self, definition_json: dict[str, Any]) -> list[StorageContainer]:
        """Gets the containers for this storage account from the definition json

        Args:
            definition_json (dict): The definition json

        Returns:
            list[StorageContainer]: The list associated containers
        """

        if self.CONTAINERS_KEY in definition_json and type(definition_json[self.CONTAINERS_KEY]) is list:
            containers: list[StorageContainer] = []
            
            for container_definition in definition_json[self.CONTAINERS_KEY]:
                container: StorageContainer = StorageContainer(container_definition, self)
                containers.append(container)
            
            return containers
        
        # If no containers are defined, return an empty list
        return []
    
    def __get_tables(self, definition_json: dict[str, Any]) -> list[StorageTable]:
        """Gets the tables for this storage account from the definition json

        Args:
            definition_json (dict): The definition json

        Returns:
            list[StorageTable]: The list associated tables
        """

        if self.TABLES_KEY in definition_json and type(definition_json[self.TABLES_KEY]) is list:
            tables: list[StorageTable] = []
            
            for table_definition in definition_json[self.TABLES_KEY]:
                table: StorageTable = StorageTable(table_definition, self)
                tables.append(table)
            
            return tables
        
        # If no tables are defined, return an empty list
        return []

    @property
    def name(self) -> str:
        """The name of the storage account"""

        return self.__name
    
    @property
    def resource_group(self) -> ResourceGroup:
        """The resource group of the storage account"""

        return self.__resource_group

    @property
    def web_apps(self) -> list[WebApp]: # type: ignore
        """The name of the webapp that uses this storage account"""

        return self.__web_apps
    
    @property
    def category(self) -> str | None:
        """The category of the storage account"""

        return self.__category_name

    @property
    def key(self) -> str:
        """The key of the storage account
        
        Note, this comes from a seperate "storage accounts" file rather than the "resources" file
        """
        return self.__key
    
    @key.setter
    def key(self, value: str):
        self.__key = value
    
    @property
    def connection_string(self) -> str:
        """The connection string for the storage account
        
        Note, this comes from a seperate "storage accounts" file rather than the "resources" file
        """
        return self.__connection_string
    
    @connection_string.setter
    def connection_string(self, value: str):
        self.__connection_string = value

    @property
    def containers(self) -> list[StorageContainer]: # type: ignore
        """The storage container for the storage account"""
        
        return self.__containers
    
    def add_container(self, container: StorageContainer): # type: ignore
        self.__containers.append(container)

    def has_container(self, container_name: str) -> bool:
        for container in self.containers:
            if container.name == container_name:
                return True
        
        return False
    
    def get_container(self, container_name: str) -> StorageContainer | None:
        for container in self.containers:
            if container.name == container_name:
                return container
        
        return None
    
    def remove_container(self, container: StorageContainer): # type: ignore
        for i in range(len(self.containers)):
            if self.containers[i].name == container.name:
                del self.containers[i]
                break
    
    @property
    def tables(self) -> list[StorageTable]: # type: ignore
        return self.__tables
    
    def add_table(self, table: StorageTable): # type: ignore
        self.__tables.append(table)

    def has_table(self, table_name: str) -> bool:
        for table in self.tables:
            if table.name == table_name:
                return True
        
        return False
    
    def get_table(self, table_name: str) -> StorageTable | None:
        for table in self.tables:
            if table.name == table_name:
                return table
        
        return None
    
    def remove_table(self, table: StorageTable): # type: ignore
        for i in range(len(self.tables)):
            if self.tables[i].name == table.name:
                del self.tables[i]
                break
    
    def to_json(self) -> dict[str, Any]:
        return_obj = {
            self.NAME_KEY: self.name,
            self.RESOURCE_GROUP_KEY: self.resource_group.name,
            self.WEB_APPS_KEY: [web_app.name for web_app in self.web_apps]
        }

        if self.key != '':
            return_obj["key"] = self.key
        
        if self.connection_string != '':
            return_obj["connectionString"] = self.connection_string
        
        if len(self.containers) > 0:
            return_obj[self.CONTAINERS_KEY] = [container.to_json() for container in self.containers]
        
        if len(self.tables) > 0:
            return_obj[self.TABLES_KEY] = [table.to_json() for table in self.tables]
        
        return return_obj


from typing import Any

from .interfaces.IContainerRegistry import IContainerRegistry
from .interfaces.IAzureResources import IAzureResources

from .ResourceGroup import ResourceGroup
from .WebApp import WebApp

class ContainerRegistry(IContainerRegistry):
    def __init__(self, definition_json: dict[str, str | list[str]], azure_resources: IAzureResources, category_name: str | None = None):
        self.__definition_json = definition_json

        self.__name = str(definition_json[self.NAME_KEY])
        self.__resource_group: ResourceGroup | None = azure_resources.get_resource_group(str(definition_json[self.RESOURCE_GROUP_KEY])) # type: ignore
        self.__category = category_name
        
        self.__web_apps: list[WebApp] = []
        if self.WEB_APPS_KEY in definition_json:
            for web_app_definition in list(definition_json[self.WEB_APPS_KEY]):
                web_app: WebApp = azure_resources.get_web_app(web_app_definition) # type: ignore
                self.__web_apps.append(web_app)

    @property
    def name(self) -> str:
        """The name of the container registry"""

        return self.__name
    
    @property
    def resource_group(self) -> ResourceGroup | None: # type: ignore
        """The resource group the container registry is in"""
        
        return self.__resource_group
    
    @property
    def category(self) -> str | None:
        """The category the container registry is in"""
        
        return self.__category

    @property
    def web_apps(self) -> list[WebApp]: # type: ignore
        """The webapps that use the container registry"""
        
        return self.__web_apps
    
    def to_json(self) -> dict[str, Any]:
        return {
            self.NAME_KEY: self.name,
            self.RESOURCE_GROUP_KEY: self.resource_group.name if self.resource_group else None,
            self.WEB_APPS_KEY: [web_app.name for web_app in self.web_apps]
        }


from abc import ABCMeta, abstractmethod

from .IResourceGroup import IResourceGroup

class IResource:
    """An abstract class that defines properties for any kind of resource (namely their name and resource group)"""

    __metaclass__ = ABCMeta

    NAME_KEY = "name"
    """The key for the name of the resource in the definition JSON"""
    RESOURCE_GROUP_KEY = "resourceGroup"
    """The key for the resource group of the resource in the definition JSON"""

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def resource_group(self) -> IResourceGroup:
        pass


from abc import ABCMeta, abstractmethod

class IAppRegistration:
    __metaclass__ = ABCMeta

    NAME_KEY = "name"
    """The key for the name of the app registration in the definition JSON"""

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the app registration"""
        pass

    @property
    @abstractmethod
    def app_id(self) -> str:
        """The app id of the app registration"""
        pass

    @app_id.setter
    @abstractmethod
    def app_id(self, value: str):
        pass

    @property
    @abstractmethod
    def app_secret(self) -> str:
        """The app secret of associated with the app registration"""
        pass

    @app_secret.setter
    @abstractmethod
    def app_secret(self, value: str):
        pass

    @property
    @abstractmethod
    def service_principal_id(self) -> str:
        """The id of the ervice principal associated with the app registration"""
        pass

    @service_principal_id.setter
    @abstractmethod
    def service_principal_id(self, value: str):
        pass


from abc import ABCMeta, abstractmethod

from .ICategorizableResource import ICategorizableResource

class IWebApp(ICategorizableResource):
    __metaclass__ = ABCMeta
    
    APP_SERVICE_PLAN_KEY = "appServicePlan"
    """The key for the app service plan of the webapp in the definition JSON"""
    INSIGHTS_KEY = "insights"
    """The key for the insights of the webapp in the definition JSON"""

    @property
    @abstractmethod
    def app_service_plan(self) -> str:
        """The app service plan of the webapp"""
        pass
    
    @property
    @abstractmethod
    def insights(self) -> bool:
        """If insights is enabled for the webapp"""
        pass
    


from abc import ABCMeta, abstractmethod

class IVarNameStorageContainer:
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def category(self) -> str | None:
        """The BA infrastructure category of the storage container"""
        pass
    
    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the storage container"""
        pass
    
    @property
    @abstractmethod
    def var_name(self) -> str:
        """The variable name for the storage container"""
        
        pass


from abc import ABCMeta, abstractmethod

from .IVarName import IVarName

class IStorageContainer:
    """An abstract representation of the internal representation of an Azure Storage Container"""

    __metaclass__ = ABCMeta
    
    NAME_KEY = "name"
    """The key for the name of the storage container in the definition JSON"""
    VAR_NAME_KEY = "varName"
    """The key for the variable name of the storage container in the definition JSON"""

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the storage container"""

        pass
    
    @property
    @abstractmethod
    def var_name_object(self) -> IVarName:
        """The variable name (object) for the storage container"""

        pass
    
    @property
    @abstractmethod
    def var_name(self) -> str:
        """The variable name for the storage container"""

        pass


from typing import Any

from .interfaces.IRedisCache import IRedisCache
from .interfaces.IAzureResources import IAzureResources

from .ResourceGroup import ResourceGroup
from .WebApp import WebApp

class RedisCache(IRedisCache):
    """Represents a redis cache in Azure"""
    
    def __init__(self, definition_json: dict[str, Any], azure_resources: IAzureResources, category_name: str | None = None):
        self.__definition_json = definition_json

        self.__name = definition_json[self.NAME_KEY]
        self.__category_name = category_name
        self.__resource_group = self.__get_resource_group(definition_json, azure_resources)
        self.__web_apps = self.__get_web_apps(definition_json, azure_resources)
    
    def __get_resource_group(self, definition_json: dict[str, Any], azure_resources: IAzureResources) -> ResourceGroup:
        """Gets the resource group from the definition json

        Args:
            definition_json (dict): The definition json
            azure_resources (IAzureResources): The azure resources (used to get the resource group)

        Raises:
            Exception: If the resource group is not found

        Returns:
            ResourceGroup: The resource group
        """

        resource_group: ResourceGroup | None = azure_resources.get_resource_group(definition_json[self.RESOURCE_GROUP_KEY]) if self.RESOURCE_GROUP_KEY in definition_json else None # type: ignore
        if resource_group is None and len(self.__web_apps) > 0:
            web_app_rgs: list[ResourceGroup] = [web_app.resource_group for web_app in self.__web_apps]
            if len(web_app_rgs) == 1:
                resource_group = web_app_rgs[0]
        
        if resource_group is None:
            raise Exception(f"Could not find resource group for database {self.name}")
    
        return resource_group
    
    def __get_web_apps(self, definition_json: dict[str, Any], azure_resources: IAzureResources) -> list[WebApp]:
        """Gets the associated web apps from the definition json

        Args:
            definition_json (dict): The definition json
            azure_resources (IAzureResources): The azure resources (used to get the web apps)

        Returns:
            list[WebApp]: The list associated web apps
        """

        if self.WEB_APPS_KEY in definition_json and type(definition_json[self.WEB_APPS_KEY]) is list:
            web_apps: list[WebApp] = []
            
            for web_app in definition_json[self.WEB_APPS_KEY]:
                web_app_object: WebApp = azure_resources.get_web_app(str(web_app)) # type: ignore
                web_apps.append(web_app_object)
            
            return web_apps
        else:
            return []
    
    @property
    def name(self) -> str:
        """The name of the redis cache"""

        return self.__name
    
    @property
    def resource_group(self) -> ResourceGroup:
        """The resource group of the redis cache"""

        return self.__resource_group

    @property
    def category_name(self) -> str | None:
        """The category name of the redis cache"""

        return self.__category_name
    
    @property
    def web_apps(self) -> list[WebApp]: # type: ignore
        """The associated web apps"""

        return self.__web_apps
    
    def to_json(self) -> dict[str, Any]:
        return {
            self.NAME_KEY: self.name,
            self.RESOURCE_GROUP_KEY: self.resource_group.name,
            self.WEB_APPS_KEY: [web_app.name for web_app in self.web_apps]
        }
    



from abc import ABCMeta, abstractmethod

from typing import Any

from .IVarNameStorageTable import IVarNameStorageTable
from .IVarNameStorageContainer import IVarNameStorageContainer

class IVarName:
    __metacalss__ = ABCMeta

    @abstractmethod
    def handle_category_prefix(self, var_name_props: dict[Any, Any], category_name: str) -> str:
        pass
    
    @abstractmethod
    def handle_replacements(self, var_name_props: dict[Any, Any], var_name_so_far: str) -> str:
        """Perform the replacements specified in the variable name's property to the encapsulating object's name

        Args:
            var_name_props (dict): The variable name's properties (where the replacements would be set, if set)
            var_name_so_far (str): The variable name so far (the name of the encapsulating object with any prefixes if set)

        Returns:
            str: The variable name with replacements filled in, if set. Otherwise the variable name as is
        """
        pass

    @abstractmethod
    def handle_suffix(self, var_name_props: dict[Any, Any], var_name_so_far: str) -> str:
        """Provide the suffix if set

        Args:
            var_name_props (dict): The variable name properties (where the suffix would be set, if set)
            encapsulating_obj (dict): The encapsulating object (the object the variable name is for)

        Returns:
            str: The suffix to use, if set. Otherwise an empty string
        """
        pass
    
    @abstractmethod
    def handle_dictionary_value(self, var_name_props: dict[Any, Any], encapsulating_obj: IVarNameStorageTable | IVarNameStorageContainer) -> str:
        pass
    
    @abstractmethod
    def get(self, encapsulating_obj: IVarNameStorageTable | IVarNameStorageContainer) -> str:
        pass


from abc import ABCMeta, abstractmethod

from .ICategorizableResource import ICategorizableResource
from .IWebAppAssociatedResource import IWebAppAssociatedResource

class IContainerRegistry(ICategorizableResource, IWebAppAssociatedResource):
    """An abstract class that defines properties for any kind of Azure Container Registry"""

    __metaclass__ = ABCMeta